@@ -1,5 +1,5 @@
-#!/usr/bin/env ruby
-#
+# encoding: utf-8
+
 # = plist
 #
 # Copyright 2006-2010 Ben Bleything and Patrick May
@@ -11,42 +11,48 @@
 # === Load a plist file
 # This is the main point of the library:
 #
-#   r = Plist::parse_xml( filename_or_xml )
+#   r = Plist.parse_xml(filename_or_xml)
 module Plist
-# Note that I don't use these two elements much:
-#
-#  + Date elements are returned as DateTime objects.
-#  + Data elements are implemented as Tempfiles
-#
-# Plist::parse_xml will blow up if it encounters a Date element.
-# If you encounter such an error, or if you have a Date element which
-# can't be parsed into a Time object, please send your plist file to
-# plist@hexane.org so that I can implement the proper support.
-  def Plist::parse_xml( filename_or_xml )
+  # Raised when an element is not implemented
+  class UnimplementedElementError < RuntimeError; end
+
+  # Note that I don't use these two elements much:
+  #
+  #  + Date elements are returned as DateTime objects.
+  #  + Data elements are implemented as Tempfiles
+  #
+  # Plist.parse_xml will blow up if it encounters a Date element.
+  # If you encounter such an error, or if you have a Date element which
+  # can't be parsed into a Time object, please create an issue
+  # attaching your plist file at https://github.com/patsplat/plist/issues
+  # so folks can implement the proper support.
+  def self.parse_xml(filename_or_xml)
     listener = Listener.new
-    #parser = REXML::Parsers::StreamParser.new(File.new(filename), listener)
+    # parser = REXML::Parsers::StreamParser.new(File.new(filename), listener)
     parser = StreamParser.new(filename_or_xml, listener)
     parser.parse
     listener.result
   end
 
   class Listener
-    #include REXML::StreamListener
+    # include REXML::StreamListener
 
     attr_accessor :result, :open
 
     def initialize
       @result = nil
-      @open   = Array.new
+      @open   = []
     end
 
-
     def tag_start(name, attributes)
-      @open.push PTag::mappings[name].new
+      @open.push PTag.mappings[name].new
     end
 
-    def text( contents )
-      @open.last.text = contents if @open.last
+    def text(contents)
+      if @open.last
+        @open.last.text ||= ''
+        @open.last.text.concat(contents)
+      end
     end
 
     def tag_end(name)
@@ -60,11 +66,11 @@ module Plist
   end
 
   class StreamParser
-    def initialize( plist_data_or_file, listener )
+    def initialize(plist_data_or_file, listener)
       if plist_data_or_file.respond_to? :read
         @xml = plist_data_or_file.read
       elsif File.exist? plist_data_or_file
-        @xml = File.read( plist_data_or_file )
+        @xml = File.read(plist_data_or_file)
       else
         @xml = plist_data_or_file
       end
@@ -72,26 +78,35 @@ module Plist
       @listener = listener
     end
 
-    TEXT       = /([^<]+)/
-    XMLDECL_PATTERN = /<\?xml\s+(.*?)\?>*/um
-    DOCTYPE_PATTERN = /\s*<!DOCTYPE\s+(.*?)(\[|>)/um
-    COMMENT_START = /\A<!--/u
-    COMMENT_END = /.*?-->/um
-
+    TEXT = /([^<]+)/
+    CDATA = /<!\[CDATA\[(.*?)\]\]>/
+    XMLDECL_PATTERN = /<\?xml\s+(.*?)\?>*/m
+    DOCTYPE_PATTERN = /\s*<!DOCTYPE\s+(.*?)(\[|>)/m
+    COMMENT_START = /\A<!--/
+    COMMENT_END = /.*?-->/m
+    UNIMPLEMENTED_ERROR = 'Unimplemented element. ' \
+      'Consider reporting via https://github.com/patsplat/plist/issues'
 
     def parse
-      plist_tags = PTag::mappings.keys.join('|')
+      plist_tags = PTag.mappings.keys.join('|')
       start_tag  = /<(#{plist_tags})([^>]*)>/i
       end_tag    = /<\/(#{plist_tags})[^>]*>/i
 
       require 'strscan'
 
-      @scanner = StringScanner.new( @xml )
+      @scanner = StringScanner.new(@xml)
       until @scanner.eos?
         if @scanner.scan(COMMENT_START)
           @scanner.scan(COMMENT_END)
         elsif @scanner.scan(XMLDECL_PATTERN)
+          encoding = parse_encoding_from_xml_declaration(@scanner[1])
+          next if encoding.nil?
+
+          # use the specified encoding for the rest of the file
+          next unless String.method_defined?(:force_encoding)
+          @scanner.string = @scanner.rest.force_encoding(encoding)
         elsif @scanner.scan(DOCTYPE_PATTERN)
+          next
         elsif @scanner.scan(start_tag)
           @listener.tag_start(@scanner[1], nil)
           if (@scanner[2] =~ /\/$/)
@@ -99,32 +114,49 @@ module Plist
           end
         elsif @scanner.scan(TEXT)
           @listener.text(@scanner[1])
+        elsif @scanner.scan(CDATA)
+          @listener.text(@scanner[1])
         elsif @scanner.scan(end_tag)
           @listener.tag_end(@scanner[1])
         else
-          raise "Unimplemented element"
+          raise UnimplementedElementError.new(UNIMPLEMENTED_ERROR)
         end
       end
     end
+
+    private
+
+    def parse_encoding_from_xml_declaration(xml_declaration)
+      return unless defined?(Encoding)
+
+      xml_encoding = xml_declaration.match(/(?:\A|\s)encoding=(?:"(.*?)"|'(.*?)')(?:\s|\Z)/)
+
+      return if xml_encoding.nil?
+
+      begin
+        Encoding.find(xml_encoding[1])
+      rescue ArgumentError
+        nil
+      end
+    end
   end
 
   class PTag
-    @@mappings = { }
-    def PTag::mappings
-      @@mappings
+    def self.mappings
+      @mappings ||= {}
     end
 
-    def PTag::inherited( sub_class )
+    def self.inherited(sub_class)
       key = sub_class.to_s.downcase
-      key.gsub!(/^plist::/, '' )
+      key.gsub!(/^plist::/, '')
       key.gsub!(/^p/, '')  unless key == "plist"
 
-      @@mappings[key] = sub_class
+      mappings[key] = sub_class
     end
 
     attr_accessor :text, :children
     def initialize
-      @children = Array.new
+      @children = []
     end
 
     def to_ruby
@@ -140,7 +172,7 @@ module Plist
 
   class PDict < PTag
     def to_ruby
-      dict = Hash.new
+      dict = {}
       key = nil
 
       children.each do |c|
@@ -158,13 +190,13 @@ module Plist
 
   class PKey < PTag
     def to_ruby
-      CGI::unescapeHTML(text || '')
+      CGI.unescapeHTML(text || '')
     end
   end
 
   class PString < PTag
     def to_ruby
-      CGI::unescapeHTML(text || '')
+      CGI.unescapeHTML(text || '')
     end
   end
 
@@ -213,7 +245,7 @@ module Plist
       data = Base64.decode64(text.gsub(/\s+/, '')) unless text.nil?
       begin
         return Marshal.load(data)
-      rescue Exception => e
+      rescue Exception
         io = StringIO.new
         io.write data
         io.rewind
